/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/
/* $XConsortium: mifillsetsZ.c,v 1.2 94/01/07 15:14:55 takamori Exp $ */
#include "X.h"
#include "gcstruct.h"
#include "miscanfillZ.h"
#include "mispansZ.h"
#include "mipolyZ.h"
#include "pixmap.h"
#include "ddpex.h"
#include "ddpex3.h"

static int getFillAreaYBounds();

#define _DONT_IGNORE_PROC_
#define _DEBUG_ON_

extern void miZloadAET(), miZcomputeWAET(), miZFreeStorage();

/*
 *
 *     Written by Daiji Takamori;  Sep 1993
 *
 *     Routines to build and fill a polygon set.  Two fill rules are
 *     supported: frWINDING and frEVENODD.
 *
 *     See mipoly.h for a complete description of the algorithm.
 */

Bool
miZGetPolySpans(spans, dst, pgc, shape, mode, count, pPts)
    ZSpans         *spans;
    DrawablePtr   dst;
    GCPtr         pgc;
    int           shape, mode;
    register int  count;
    ddCoord3DS    *pPts;
{
    int			i;
    register int	xorg, yorg;
    register ddCoord3DS *ppt;

    if (count == 0)
	return;

    ppt = pPts;
    if (pgc->miTranslate)
    {
	xorg = dst->x;
	yorg = dst->y;

        if (mode == CoordModeOrigin) 
        {
	        for (i = 0; i<count; i++) 
                {    
	            ppt->x += xorg;
	            ppt++->y += yorg;
	        }
        }
        else 
        {
	    ppt->x += xorg;
	    ppt++->y += yorg;
	    for (i = 1; i<count; i++) 
            {
	        ppt->x += (ppt-1)->x;
	        ppt->y += (ppt-1)->y;
	        ppt++;
	    }
        }
    }
    else
    {
	if (mode == CoordModePrevious)
        {
	    ppt++;
	    for (i = 1; i<count; i++) 
            {
	        ppt->x += (ppt-1)->x;
	        ppt->y += (ppt-1)->y;
	        ppt++;
	    }
        }
    }
    if (shape == Convex)
	miZGetConvexPolySpans(spans, pgc, count, pPts);
    else
	miZGetGeneralPolySpans(spans, pgc, count, pPts);
}

Bool
miZAppendPolygonSet(polyset, shape, count, ptsIn)
    int shape, count;
    ddCoord3DS *ptsIn;
{
}

Bool
miZGetConvexPolySpans(spans, pgc, count, ptsIn)
    ZSpans        *spans;
    GCPtr         pgc;
    register int  count;
    ddCoord3DS    *ptsIn;
{
    register int xl, xr;        /* x vals of left and right edges   */
    register int zl, zr;        /* z vals of left and right edges   */
    register int dl, dr;        /* decision variables               */
    register int dzl, dzr;      /* decision variables for z increment */
    register int ml, m1l;       /* left edge slope and slope+1      */
    register int mzl, mz1l;     /* left edge z-slope and z-slope+1  */
    int mr, m1r;                /* right edge slope and slope+1     */
    int mzr, mz1r;              /* right edge z-slope and z-slope+1 */
    int incr1l, incr2l;         /* left edge error increments       */
    int incrz1l, incrz2l;       /* left edge z error increments     */
    int incr1r, incr2r;         /* right edge error increments      */
    int incrz1r, incrz2r;       /* right edge z error increments    */
    int dy;                     /* delta y                          */
    int y;                      /* current scanline                 */
    int left, right;            /* indices to first endpoints       */
    int i;                      /* loop counter                     */
    int nextleft, nextright;    /* indices to second endpoints      */
    ddCoord3DS *ptsOut;         /* output buffer                    */
    int *width, *depth;         /* output buffer                    */
    int imin;                   /* index of smallest vertex (in y)  */
    int ymin;                   /* y-extents of polygon             */
    int ymax;

    /*
     *  find leftx, bottomy, rightx, topy, and the index
     *  of bottomy. Also translate the points.
     */
    imin = getFillAreaYBounds(ptsIn, count, &ymin, &ymax);

    dy = ymax - ymin + 1;
    if ((count < 3) || (dy < 0))
	return(TRUE);

    spans->count = 0;
    ptsOut = spans->points = (ddCoord3DS *) xalloc(sizeof(ddCoord3DS) * dy);
    width = spans->widths = (int *) xalloc(sizeof(int) * dy);
    depth = spans->depths = (int *) xalloc(sizeof(int) * dy);
    if(!ptsOut || !width || !depth)
    {
	if (depth) xfree(depth);
	if (width) xfree(width);
	if (ptsOut) xfree(ptsOut);
	return(FALSE);
    }
    nextleft = nextright = imin;
    y = ptsIn[nextleft].y;

    /*
     *  loop through all edges of the polygon
     */
    do {
        /*
         *  add a left edge if we need to
         */
        if (ptsIn[nextleft].y == y) {
            left = nextleft;

            /*
             *  find the next edge, considering the end
             *  conditions of the array.
             */
            nextleft++;
            if (nextleft >= count)
                nextleft = 0;

            /*
             *  now compute all of the random information
             *  needed to run the iterative algorithm.
             */
            ZBRESINITPGON(ptsIn[nextleft].y-ptsIn[left].y,
                         ptsIn[left].x,ptsIn[nextleft].x,
                         ptsIn[left].z,ptsIn[nextleft].z,
                         xl, zl, dl, dzl, ml, m1l, mzl, mz1l,
			 incr1l, incr2l, incrz1l, incrz2l);
        }

        /*
         *  add a right edge if we need to
         */
        if (ptsIn[nextright].y == y) {
            right = nextright;

            /*
             *  find the next edge, considering the end
             *  conditions of the array.
             */
            nextright--;
            if (nextright < 0)
                nextright = count-1;

            /*
             *  now compute all of the random information
             *  needed to run the iterative algorithm.
             */
            ZBRESINITPGON(ptsIn[nextright].y-ptsIn[right].y,
                         ptsIn[right].x,ptsIn[nextright].x,
                         ptsIn[right].z,ptsIn[nextright].z,
                         xr, zr, dr, dzr, mr, m1r, mzr, mz1r,
			 incr1r, incr2r, incrz1r, incrz2r);
        }

        /*
         *  generate scans to fill while we still have
         *  a right edge as well as a left edge.
         */
        i = min(ptsIn[nextleft].y, ptsIn[nextright].y) - y;
	/* in case we're called with non-convex polygon */
	if(i < 0)
        {
	    xfree(spans->depths);
	    xfree(spans->widths);
	    xfree(spans->points);
	    return(TRUE);
	}
        while (i-- > 0) 
        {
            ptsOut->y = y;

            /*
             *  reverse the edges if necessary
             */
            if (xl < xr) 
            {
                *(depth++) = zr - zl;
                *(width++) = xr - xl;
                ptsOut->x = xl;
                (ptsOut++)->z = zl;
            }
            else
            {
                *(depth++) = zl - zr;
                *(width++) = xl - xr;
                ptsOut->x = xr;
                (ptsOut++)->z = zr;
            }
            y++;

            /* increment down the edges */
            ZBRESINCRPGON(dl, xl, ml, m1l, incr1l, incr2l);
            ZBRESINCRPGON(dzl, zl, mzl, mz1l, incrz1l, incrz2l);
            ZBRESINCRPGON(dr, xr, mr, m1r, incr1r, incr2r);
            ZBRESINCRPGON(dzr, zr, mzr, mz1r, incrz1r, incrz2r);
        }
    }  while (y != ymax);

    /*  The spans set is complete (this is a convex polygon) */
    spans->count = ptsOut-spans->points;
    return(TRUE);
}

/*
 *     Find the index of the point with the smallest y.
 */
static
int
getFillAreaYBounds(pts, n, by, ty)
    ddCoord3DS *pts;
    int n;
    int *by, *ty;
{
    register ddCoord3DS *ptMin;
    int ymin, ymax;
    ddCoord3DS *ptsStart = pts;

    ptMin = pts;
    ymin = ymax = (pts++)->y;

    while (--n > 0) {
        if (pts->y < ymin)
	{
            ptMin = pts;
            ymin = pts->y;
        }
	if(pts->y > ymax)
            ymax = pts->y;

        pts++;
    }

    *by = ymin;
    *ty = ymax;
    return(ptMin-ptsStart);
}

Bool
miZGetGeneralPolySpans(spans, pgc, count, ptsIn)
    ZSpans        *spans;
    GCPtr         pgc;
    register int  count;
    ddCoord3DS    *ptsIn;
{
    register ZEdgeTableEntry *pAET;  /* the Active Edge Table   */
    register int y;                 /* the current scanline    */
    register int nPts = 0;          /* number of pts in buffer */
    register ZEdgeTableEntry *pWETE; /* Winding Edge Table      */
    register ZScanLineList *pSLL;    /* Current ScanLineList    */
    ddCoord3DS *ptsOut;              /* the output buffers */
    int *width, *depth;
    ZEdgeTableEntry *pPrevAET;       /* previous AET entry      */
    ZEdgeTable ET;                   /* Edge Table header node  */
    ZEdgeTableEntry AET;             /* Active ET header node   */
    ZEdgeTableEntry *pETEs;          /* Edge Table Entries buff */
    ZScanLineListBlock SLLBlock;     /* header for ScanLineList */
    int fixWAET = 0;

    if (count < 3)
	return(TRUE);

    spans->count = NUMPTSTOBUFFER;
    ptsOut = spans->points = (ddCoord3DS *) xalloc(sizeof(ddCoord3DS) * NUMPTSTOBUFFER);
    width = spans->widths = (int *) xalloc(sizeof(int) * NUMPTSTOBUFFER);
    depth = spans->depths = (int *) xalloc(sizeof(int) * NUMPTSTOBUFFER);
    if(!ptsOut || !width || !depth)
    {
	if (depth) xfree(depth);
	if (width) xfree(width);
	if (ptsOut) xfree(ptsOut);
	DEALLOCATE_LOCAL(pETEs);
	return(FALSE);
    }
#ifdef _IGNORE_PROC_
return(TRUE);
#endif
    if(!(pETEs = (ZEdgeTableEntry *)
	           ALLOCATE_LOCAL(sizeof(ZEdgeTableEntry) * count)))
	return(FALSE);
    if (!miZCreateETandAET(count, ptsIn, &ET, &AET, pETEs, &SLLBlock))
    {
	DEALLOCATE_LOCAL(pETEs);
	return(FALSE);
    }
    pSLL = ET.scanlines.next;
    if (pgc->fillRule == EvenOddRule) 
    {
        /*
         *  for each scanline
         */
        for (y = ET.ymin; y < ET.ymax; y++) 
        {
            /*
             *  Add a new edge to the active edge table when we
             *  get to the next edge.
             */
            if (pSLL && y == pSLL->scanline) 
            {
                miZloadAET(&AET, pSLL->edgelist);
                pSLL = pSLL->next;
            }
            pPrevAET = &AET;
            pAET = AET.next;

            /*
             *  for each active edge
             */
            while (pAET) 
            {
                ptsOut->x = pAET->bres.minor;
                ptsOut->z = pAET->bres.minorz;
		ptsOut++->y = y;
                *width++ = pAET->next->bres.minor - pAET->bres.minor;
                *depth++ = pAET->next->bres.minorz - pAET->bres.minorz;
                nPts++;

                /*
                 *  expand the buffer when its full
                 */
                if (nPts == NUMPTSTOBUFFER) 
		{
		    int         index;

		    spans->count += NUMPTSTOBUFFER;
		    index = ptsOut - spans->points;
		    spans->points = (ddCoord3DS *) xrealloc (spans->points, 
						     spans->count * sizeof (ddCoord3DS));
		    spans->widths = (int *) xrealloc (spans->widths, 
					       spans->count * sizeof (int));
		    spans->depths = (int *) xrealloc (spans->depths, 
					       spans->count * sizeof (int));
		    if (!spans->points || !spans->widths || !spans->depths) 
		    {
		        if (spans->depths) xfree(spans->depths);
			if (spans->widths) xfree(spans->widths);
			if (spans->points) xfree(spans->points);
			DEALLOCATE_LOCAL(pETEs);
			return(FALSE);
		    }
		    ptsOut = spans->points + index;
		    width = spans->widths + index;
		    depth = spans->depths + index;
                    nPts = 0;
                }
                ZEVALUATEEDGEEVENODD(pAET, pPrevAET, y);
                ZEVALUATEEDGEEVENODD(pAET, pPrevAET, y);
            }
            miZInsertionSort(&AET);
        }
    }
    else      /* default to WindingNumber */
    {
        /*
         *  for each scanline
         */
        for (y = ET.ymin; y < ET.ymax; y++) 
        {
            /*
             *  Add a new edge to the active edge table when we
             *  get to the next edge.
             */
            if (pSLL && y == pSLL->scanline) 
            {
                miZloadAET(&AET, pSLL->edgelist);
                miZcomputeWAET(&AET);
                pSLL = pSLL->next;
            }
            pPrevAET = &AET;
            pAET = AET.next;
            pWETE = pAET;

            /*
             *  for each active edge
             */
            while (pAET) 
            {
                /*
                 *  if the next edge in the active edge table is
                 *  also the next edge in the winding active edge
                 *  table.
                 */
                if (pWETE == pAET) 
                {
                    ptsOut->x = pAET->bres.minor;
		    ptsOut++->y = y;
                    ptsOut->z = pAET->bres.minorz;
                    *width++ = pAET->nextWETE->bres.minor - pAET->bres.minor;
                    *depth++ = pAET->nextWETE->bres.minorz - pAET->bres.minorz;
                    nPts++;

                    /*
                     *  expand the buffer if necessary
                     */
                    if (nPts == NUMPTSTOBUFFER) 
                    {
		        ddCoord3DS  *newPt;
			int         *newWid, *newDep;
			
			spans->count += NUMPTSTOBUFFER;
			newPt = (ddCoord3DS *) xrealloc (spans->points, 
							 spans->count * sizeof (ddCoord3DS));
			newWid = (int *) xrealloc (spans->widths, 
						   spans->count * sizeof (int));
			newDep = (int *) xrealloc (spans->depths, 
						   spans->count * sizeof (int));
			if (!newPt || !newWid || !newDep)
			{
			    if (depth) xfree(depth);
			    if (width) xfree(width);
			    if (ptsOut) xfree(ptsOut);
			    DEALLOCATE_LOCAL(pETEs);
			    return(FALSE);
			}
			ptsOut = newPt + (ptsOut - spans->points);
			width = newWid + (width - spans->widths);
			depth = newDep + (depth - spans->depths);
			spans->points = newPt;
			spans->widths = newWid;
			spans->depths = newDep;
			nPts = 0;
                    }

                    pWETE = pWETE->nextWETE;
                    while (pWETE != pAET)
                        ZEVALUATEEDGEWINDING(pAET, pPrevAET, y, fixWAET);
                    pWETE = pWETE->nextWETE;
                }
                ZEVALUATEEDGEWINDING(pAET, pPrevAET, y, fixWAET);
            }

            /*
             *  reevaluate the Winding active edge table if we
             *  just had to resort it or if we just exited an edge.
             */
            if (miZInsertionSort(&AET) || fixWAET) 
            {
                miZcomputeWAET(&AET);
                fixWAET = 0;
            }
        }
    }

    /*
     *     Get any spans that we missed by buffering
     */
    spans->count -= NUMPTSTOBUFFER - nPts;
    DEALLOCATE_LOCAL(pETEs);
    miZFreeStorage(SLLBlock.next);
    return(TRUE);
}
